@@ -32,7 +32,7 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1997 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dproc.c,v 1.21 2008/04/15 13:32:26 abe Exp $";
+static char *rcsid = "$Id: dproc.c,v 1.30 2018/02/14 14:26:38 abe Exp $";
 #endif
 
 #include "lsof.h"
@@ -44,7 +44,17 @@ static char *rcsid = "$Id: dproc.c,v 1.21 2008/04/15 13:32:26 abe Exp $";
 
 #define	FDINFO_FLAGS		1	/* fdinfo flags available */
 #define	FDINFO_POS		2	/* fdinfo position available */
-#define FDINFO_ALL		(FDINFO_FLAGS | FDINFO_POS)
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+#define FDINFO_TTY_INDEX	4	/* fdinfo tty-index available */
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+#define FDINFO_ALL		(FDINFO_FLAGS | FDINFO_POS | FDINFO_TTY_INDEX)
+#else   /* !(defined(HASEPTOPTS) && defined(HASPTYEPT)) */
+#define FDINFO_ALL		(FDINFO_FLAGS | FDINFO_POS )
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
 #define	LSTAT_TEST_FILE		"/"
 #define LSTAT_TEST_SEEK		1
 
@@ -60,6 +70,11 @@ static char *rcsid = "$Id: dproc.c,v 1.21 2008/04/15 13:32:26 abe Exp $";
 struct l_fdinfo {
 	int flags;			/* flags: line value */
 	off_t pos;			/* pos: line value */
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+	int tty_index;			/* pty line index */
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
 };
 
 
@@ -67,22 +82,31 @@ struct l_fdinfo {
  * Local variables
  */
 
-short Cscko, Scko;			/* socket processing options */
-struct lproc *PLp;			/* process' Lp */
+static short Cckreg;			/* conditional status of regular file
+					 * checking:
+					 *     0 = unconditionally check
+					 *     1 = conditionally check */
+static short Ckscko;			/* socket file only checking status:
+					 *     0 = none
+					 *     1 = check only socket files */
 
 
 /*
  * Local function prototypes
  */
 
-_PROTOTYPE(static int get_fdinfo,(char *p, struct l_fdinfo *fi));
-_PROTOTYPE(static int getlinksrc,(char *ln, char *src, int srcl));
+_PROTOTYPE(static MALLOC_S alloc_cbf,(MALLOC_S len, char **cbf, MALLOC_S cbfa));
+_PROTOTYPE(static int get_fdinfo,(char *p, int msk, struct l_fdinfo *fi));
+_PROTOTYPE(static int getlinksrc,(char *ln, char *src, int srcl, char **rest));
+_PROTOTYPE(static int isefsys,(char *path, char *type, int l,
+			       efsys_list_t **rep, struct lfile **lfr));
 _PROTOTYPE(static int nm2id,(char *nm, int *id, int *idl));
-_PROTOTYPE(static int read_id_stat,(int ty, char *p, int id, char **cmd,
-				    int *ppid, int *pgid));
+_PROTOTYPE(static int read_id_stat,(char *p, int id, char **cmd, int *ppid,
+				    int *pgid));
 _PROTOTYPE(static void process_proc_map,(char *p, struct stat *s, int ss));
 _PROTOTYPE(static int process_id,(char *idp, int idpl, char *cmd, UID_ARG uid,
-				  int pid, int ppid, int pgid));
+				  int pid, int ppid, int pgid, int tid,
+				  char *tcmd));
 _PROTOTYPE(static int statEx,(char *p, struct stat *s, int *ss));
  
 
@@ -143,6 +167,29 @@ enter_cntx_arg(cntx)
 #endif	/* defined(HASSELINUX) */
 
 
+/*
+ * alloc_cbf() -- allocate a command buffer
+ */
+
+static MALLOC_S
+alloc_cbf(len, cbf, cbfa)
+	MALLOC_S len;				/* required length */
+	char **cbf;				/* current buffer */
+	MALLOC_S cbfa;				/* current buffer allocation */
+{
+	if (*cbf)
+	    *cbf = (char *)realloc((MALLOC_P *)*cbf, len);
+	else
+	    *cbf = (char *)malloc(len);
+	if (!*cbf) {
+	    (void) fprintf(stderr,
+		"%s: can't allocate command %d bytes\n", Pn, (int)len);
+	     Exit(1);
+	}
+	return(len);
+}
+
+
 /*
  * gather_proc_info() -- gather process information
  */
@@ -150,20 +197,22 @@ enter_cntx_arg(cntx)
 void
 gather_proc_info()
 {
-	char *cmd;
+	char *cmd, *tcmd;
+	char cmdbuf[MAXPATHLEN];
 	struct dirent *dp;
-	struct stat sb;
-	int lwp, n, nl, pgid, pid, ppid, rv, tx;
-	static char *lwppath = (char *)NULL;
-	static int lwppathl = 0;
+	unsigned char ht, pidts;
+	int n, nl, pgid, pid, ppid, prv, rv, tid, tpgid, tppid, tx;
 	static char *path = (char *)NULL;
 	static int pathl = 0;
 	static char *pidpath = (char *)NULL;
 	static MALLOC_S pidpathl = 0;
 	static MALLOC_S pidx = 0;
 	static DIR *ps = (DIR *)NULL;
+	struct stat sb;
 	static char *taskpath = (char *)NULL;
 	static int taskpathl = 0;
+	static char *tidpath = (char *)NULL;
+	static int tidpathl = 0;
 	DIR *ts;
 	UID_ARG uid;
 
@@ -176,7 +225,7 @@ gather_proc_info()
 	    if (!(pidpath = (char *)malloc(pidpathl))) {
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for \"%s/\"<pid>\n",
-		    Pn, pidpathl, PROCFS);
+		    Pn, (int)pidpathl, PROCFS);
 		Exit(1);
 	    }
 	    (void) snpf(pidpath, pidpathl, "%s/", PROCFS);
@@ -205,32 +254,33 @@ gather_proc_info()
 
 	    /*
 	     * Selection ANDing or only network file options have been
-	     * specified, so set unconditional skipping of regular files.
+	     * specified, so set unconditional skipping of regular files
+	     * and socket file only checking.
 	     */
-		Cscko = 0;
-		Scko = 1;
+		Cckreg = 0;
+		Ckscko = 1;
 	    } else {
 
 	    /*
 	     * If ORed file selection options have been specified, or no ORed
-	     * process options have been specified, enable unconditional file
-	     * processing.
+	     * process selection options have been specified, enable
+	     * unconditional file checking and clear socket file only checking.
 	     *
 	     * If only ORed process selection options have been specified,
-	     * enable conditional file skipping.
+	     * enable conditional file skipping and socket file only checking.
 	     */
-		if ((Selflags & SELFILE) || !(Selflags & SELPROC))
-		    Cscko = Scko = 0;
+		if ((Selflags & SELFILE) || !(Selflags & SelProc))
+		    Cckreg = Ckscko = 0;
 		else
-		    Cscko = Scko = 1;
+		    Cckreg = Ckscko = 1;
 	    }
 	} else {
 
 	/*
 	 * No network file selection options were specified.  Enable
-	 * unconditional file processing.
+	 * unconditional file checking and clear socket file only checking.
 	 */
-	    Cscko = Scko = 0;
+	    Cckreg = Ckscko = 0;
 	}
 /*
  * Read /proc, looking for PID directories.  Open each one and
@@ -255,7 +305,7 @@ gather_proc_info()
 		{
 		    (void) fprintf(stderr,
 			"%s: can't allocate %d bytes for \"%s/%s/\"\n",
-			Pn, pidpathl, PROCFS, dp->d_name);
+			Pn, (int)pidpathl, PROCFS, dp->d_name);
 		    Exit(1);
 		}
 	    }
@@ -267,72 +317,106 @@ gather_proc_info()
 	    if (stat(pidpath, &sb))
 		continue;
 	    uid = (UID_ARG)sb.st_uid;
+	    ht = pidts = 0;
 	/*
-	 * Process the PID's process information.
+	 * Get the PID's command name.
 	 */
 	    (void) make_proc_path(pidpath, n, &path, &pathl, "stat");
-	    rv = read_id_stat(0, path, pid, &cmd, &ppid, &pgid);
-	    if (rv == 1)
-		continue;
-	    else if (rv == 0) {
-		(void) process_id(pidpath, n, cmd, uid, pid, ppid, pgid);
-		continue;
-	    }
+	    if ((prv = read_id_stat(path, pid, &cmd, &ppid, &pgid)) < 0) 
+		cmd = "(unknown)";
+
+#if	defined(HASTASKS)
 	/*
-	 * The process is a zombie.  Check for a non-zombie task.
+	 * Task reporting has been selected, so save the process' command
+	 * string, so that task processing won't change it in the buffer of
+	 * read_id_stat().
+	 *
+	 * Check the tasks of the process first, so that the "-p<PID> -aK"
+	 * options work properly.
 	 */
-	    (void) make_proc_path(pidpath, n, &taskpath, &taskpathl, "task");
-	    tx = n + 4;
-	    if ((ts = opendir(taskpath))) {
-
-	    /*
-	     * Process the PID's tasks (lightweight processes.)  Record the
-	     * open files of the first one whose LWP ID does not match the PID
-	     * and which is not a itself a zombie.
-	     */
-		while ((dp = readdir(ts))) {
+	    else if (!IgnTasks && (Selflags & SELTASK)) {
+		strncpy(cmdbuf, cmd, sizeof(cmdbuf) - 1);
+		cmdbuf[sizeof(cmdbuf) - 1] = '\0';
+		cmd = cmdbuf;
+		(void) make_proc_path(pidpath, n, &taskpath, &taskpathl,
+				      "task");
+		tx = n + 4;
+		if ((ts = opendir(taskpath))) {
 
 		/*
-		 * Get the LWP ID.  Skip the task if its LWP ID matches the
-		 * process PID.
+		 * Process the PID's tasks.  Record the open files of those
+		 * whose TIDs do not match the PID and which are themselves
+		 * not zombies.
 		 */
-		    if (nm2id(dp->d_name, &lwp, &nl))
-			continue;
-		    if  (lwp == pid)
-			continue;
-		/*
-		 * Check the LWP state.
-		 */
-		    if (read_id_stat(1, path, lwp, &cmd, &ppid, &pgid))
-			continue;
-		/*
-		 * Attempt to record the LWP.
-		 */
-		    if ((tx + 1 + nl + 1) > lwppathl) {
-			lwppathl = tx + 1 + n + 1 + 64;
-			if (lwppath)
-			    lwppath = (char *)realloc((MALLOC_P *)lwppath,
-						      lwppathl);
-			else
-			    lwppath = (char *)malloc((MALLOC_S)lwppathl);
-			if (!lwppath) {
-			    (void) fprintf(stderr,
-				"%s: can't allocate %d task bytes", Pn,
-				lwppathl);
-			    (void) fprintf(stderr,
-				" for \"%s/%s/\"\n", taskpath, dp->d_name);
-			    Exit(1);
+		    while ((dp = readdir(ts))) {
+
+		    /*
+		     * Get the task ID.  Skip the task if its ID matches the
+		     * process PID.
+		     */
+			if (nm2id(dp->d_name, &tid, &nl))
+			    continue;
+			if  (tid == pid) {
+			    pidts = 1;
+			    continue;
+			}
+		    /*
+		     * Form the path for the TID.
+		     */
+			if ((tx + 1 + nl + 1 + 4) > tidpathl) {
+			    tidpathl = tx + 1 + n + 1 + 4 + 64;
+			    if (tidpath)
+				tidpath = (char *)realloc((MALLOC_P *)tidpath,
+							  tidpathl);
+			    else
+				tidpath = (char *)malloc((MALLOC_S)tidpathl);
+			    if (!tidpath) {
+				(void) fprintf(stderr,
+				    "%s: can't allocate %d task bytes", Pn,
+				    tidpathl);
+				(void) fprintf(stderr, " for \"%s/%s/stat\"\n",
+				    taskpath, dp->d_name);
+				Exit(1);
+			    }
+			}
+			(void) snpf(tidpath, tidpathl, "%s/%s/stat", taskpath,
+			    dp->d_name);
+		    /*
+		     * Check the task state.
+		     */
+			rv = read_id_stat(tidpath, tid, &tcmd, &tppid,
+					  &tpgid);
+			if ((rv < 0) || (rv == 1))
+			    continue;
+		    /*
+		     * Attempt to record the task.
+		     */
+			if (!process_id(tidpath, (tx + 1 + nl+ 1), cmd, uid,
+					pid, tppid, tpgid, tid, tcmd))
+			{
+			    ht = 1;
 			}
 		    }
-		    (void) snpf(lwppath, lwppathl, "%s/%s/", taskpath,
-			dp->d_name);
-		    if (!process_id(lwppath, (tx + 1 + nl+ 1), cmd, uid, pid,
-				    ppid, pgid))
-		    {
-			break;
-		    }
+		    (void) closedir(ts);
+		}
+	    }
+#endif	/* defined(HASTASKS) */
+
+	/*
+	 * If the main process is a task and task selection has been specified
+	 * along with option ANDing, enter the main process temporarily as a
+	 * task, so that the "-aK" option set lists the main process along
+	 * with its tasks.
+	 */
+	    if ((prv >= 0) && (prv != 1)) {
+		tid = (Fand && ht && pidts && !IgnTasks && (Selflags & SELTASK))
+		    ? pid : 0;
+		if ((!process_id(pidpath, n, cmd, uid, pid, ppid, pgid, tid,
+				 (char *)NULL))
+		&&  tid)
+		{
+		    Lp->tid = 0;
 		}
-		(void) closedir(ts);
 	    }
 	}
 }
@@ -343,8 +427,10 @@ gather_proc_info()
  */
 
 static int
-get_fdinfo(p, fi)
+get_fdinfo(p, msk, fi)
 	char *p;			/* path to fdinfo file */
+	int msk;			/* mask for information type: e.g.,
+					 * the FDINFO_* definition */
 	struct l_fdinfo *fi;		/* pointer to local fdinfo values
 					 * return structure */
 {
@@ -359,6 +445,11 @@ get_fdinfo(p, fi)
  */
 	if (!fi)
 	    return(0);
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+	fi->tty_index = -1;
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
 	if (!p || !*p || !(fs = fopen(p, "r")))
 	    return(0);
 /*
@@ -379,7 +470,7 @@ get_fdinfo(p, fi)
 		||  !ep || *ep)
 		    continue;
 		fi->flags = (unsigned int)ul;
-		if ((rv |= FDINFO_FLAGS) == FDINFO_ALL)
+		if ((rv |= FDINFO_FLAGS) == msk)
 		    break;
 	    } else if (!strcmp(fp[0], "pos:")) {
 
@@ -393,6 +484,29 @@ get_fdinfo(p, fi)
 		fi->pos = (off_t)ull;
 		if ((rv |= FDINFO_POS) == FDINFO_ALL)
 		    break;
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+	    } else if (!strcmp(fp[0], "tty-index:")) {
+
+	    /*
+	     * Process a "tty-index:" line.
+	     */
+		ep = (char *)NULL;
+		if ((ul = strtoul(fp[1], &ep, 0)) == ULONG_MAX
+		||  !ep || *ep)
+		     continue;
+		fi->tty_index = (int)ul;
+		if (fi->tty_index < 0) {
+
+		/*
+		 * Oops! If integer overflow occurred, reset the field.
+		 */
+		     fi->tty_index = -1;
+		}
+		if ((rv |= FDINFO_TTY_INDEX) == msk)
+ 		    break;
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
 	    }
 	}
 	fclose(fs);
@@ -409,14 +523,19 @@ get_fdinfo(p, fi)
 
 
 static int
-getlinksrc(ln, src, srcl)
+getlinksrc(ln, src, srcl, rest)
 	char *ln;			/* link path */
 	char *src;			/* link source path return address */
 	int srcl;			/* length of src[] */
+	char **rest;			/* pointer to what follows the ':' in
+					 * the link source path (NULL if no
+					 * return requested) */
 {
 	char *cp;
 	int ll;
 
+	if (rest)
+	    *rest = (char *)NULL;
 	if ((ll = readlink(ln, src, srcl - 1)) < 1
 	||  ll >= srcl)
 	    return(-1);
@@ -426,6 +545,8 @@ getlinksrc(ln, src, srcl)
 	if ((cp = strchr(src, ':'))) {
 	    *cp = '\0';
 	    ll = strlen(src);
+	    if (rest)
+		*rest = cp + 1;
 	}
 	return(ll);
 }
@@ -472,7 +593,7 @@ initialize()
 	    if (!OffType) {
 		(void) snpf(path, sizeof(path), "%s/%d/fdinfo/%d", PROCFS,
 			    Mypid, fd);
-		if (get_fdinfo(path, &fi) & FDINFO_POS) {
+		if (get_fdinfo(path, FDINFO_POS, &fi) & FDINFO_POS) {
 		    if (fi.pos == (off_t)LSTAT_TEST_SEEK)
 			OffType = 2;
 		}
@@ -539,7 +660,7 @@ make_proc_path(pp, pl, np, nl, sf)
 	    if (!cp) {
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for %s%s\n",
-		    Pn, rl, pp, sf);
+		    Pn, (int)rl, pp, sf);
 		Exit(1);
 	    }
 	    *nl = rl;
@@ -551,6 +672,82 @@ make_proc_path(pp, pl, np, nl, sf)
 }
 
 
+/*
+ * isefsys() -- is path on a file system exempted with -e
+ *
+ * Note: alloc_lfile() must have been called in advance.
+ */
+
+static int
+isefsys(path, type, l, rep, lfr)
+	char *path;			/* path to file */
+	char *type;			/* unknown file type */
+	int l;				/* link request: 0 = report
+					 *               1 = link */
+	efsys_list_t **rep;		/* returned Efsysl pointer, if not
+					 * NULL */
+	struct lfile **lfr;		/* allocated struct lfile pointer */
+{
+	efsys_list_t *ep;
+	int ds, len;
+	struct mounts *mp;
+	char nmabuf[MAXPATHLEN + 1];
+
+	len = (int) strlen(path);
+	for (ep = Efsysl; ep; ep = ep->next) {
+
+	/*
+	 * Look for a matching exempt file system path at the beginning of
+	 * the file path.
+	 */
+	    if (ep->pathl > len)
+		continue;
+	    if (strncmp(ep->path, path, ep->pathl))
+		continue;
+	/*
+	 * If only reporting, return information as requested.
+	 */
+	    if (!l) {
+		if (rep)
+		    *rep = ep;
+		return(0);
+	    }
+	/*
+	 * Process an exempt file.
+	 */
+	    ds = 0;
+	    if ((mp = ep->mp)) {
+		if (mp->ds & SB_DEV) {
+		    Lf->dev = mp->dev;
+		    ds = Lf->dev_def = 1;
+		}
+		if (mp->ds & SB_RDEV) {
+		    Lf->rdev = mp->rdev;
+		    ds = Lf->rdev_def = 1;
+		}
+	    }
+	    if (!ds)
+		(void) enter_dev_ch("UNKNOWN");
+	    Lf->ntype = N_UNKN;
+	    (void) snpf(Lf->type, sizeof(Lf->type), "%s",
+			(type ? type : "UNKN"));
+	    (void) enter_nm(path);
+	    (void) snpf(nmabuf, sizeof(nmabuf), "(%ce %s)",
+		ep->rdlnk ? '+' : '-', ep->path);
+	    nmabuf[sizeof(nmabuf) - 1] = '\0';
+	    (void) add_nma(nmabuf, strlen(nmabuf));
+	    if (Lf->sf) {
+		if (lfr)
+		    *lfr = Lf;
+		link_lfile();
+	    } else if (lfr)
+		*lfr = (struct lfile *)NULL;
+	    return(0);
+	}
+	return(1);
+}
+
+
 /*
  * nm2id() - convert a name to an integer ID
  */
@@ -659,7 +856,7 @@ open_proc_stream(p, m, buf, sz, act)
  */
 
 static int
-process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
+process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid, tcmd)
 	char *idp;			/* pointer to ID's path */
 	int idpl;			/* pointer to ID's path length */
 	char *cmd;			/* pointer to ID's command */
@@ -667,11 +864,13 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
 	int pid;			/* ID's PID */
 	int ppid;			/* parent PID */
 	int pgid;			/* parent GID */
+	int tid;			/* task ID, if non-zero */
+	char *tcmd;			/* task command, if non-NULL) */
 {
-	int av;
+	int av = 0;
 	static char *dpath = (char *)NULL;
 	static int dpathl = 0;
-	short enls, enss, lnk, oty, pn, pss, sf;
+	short efs, enls, enss, lnk, oty, pn, pss, sf;
 	int fd, i, ls, n, ss, sv;
 	struct l_fdinfo fi;
 	DIR *fdp;
@@ -679,12 +878,14 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
 	static char *ipath = (char *)NULL;
 	static int ipathl = 0;
 	int j = 0;
+	struct lfile *lfr;
 	struct stat lsb, sb;
 	char nmabuf[MAXPATHLEN + 1], pbuf[MAXPATHLEN + 1];
 	static char *path = (char *)NULL;
 	static int pathl = 0;
 	static char *pathi = (char *)NULL;
 	static int pathil = 0;
+	char *rest;
 	int txts = 0;
 
 #if	defined(HASSELINUX)
@@ -694,174 +895,199 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
 /*
  * See if process is excluded.
  */
-	if (is_proc_excl(pid, pgid, uid, &pss, &sf)
+	if (is_proc_excl(pid, pgid, uid, &pss, &sf, tid)
 	||  is_cmd_excl(cmd, &pss, &sf))
+	{
+
+#if	defined(HASEPTOPTS)
+	    if (!FeptE)
+		return(1);
+#else	/* !defined(HASEPTOPTS) */
 	    return(1);
-/*
- * Set the socket processing options.
- */
-	if (Cscko) {
-	    if (sf & SELPROC)
-		Scko = 0;
-	    else
-		Scko = 1;
+#endif	/* defined(HASEPTOPTS) */
+
+	}
+	if (Cckreg && !FeptE) {
+
+	/*
+	 * If conditional checking of regular files is enabled, enable
+	 * socket file only checking, based on the process' selection
+	 * status.
+	 */
+	    Ckscko = (sf & SelProc) ? 0 : 1;
 	}
 	alloc_lproc(pid, pgid, ppid, uid, cmd, (int)pss, (int)sf);
 	Plf = (struct lfile *)NULL;
+
+#if	defined(HASTASKS)
+/*
+ * Enter task information.
+ */
+	Lp->tid = tid;
+	if (tid && tcmd) {
+	    if (!(Lp->tcmd = mkstrcpy(tcmd, (MALLOC_S *)NULL))) {
+		(void) fprintf(stderr,
+		    "%s: PID %d, TID %d, no space for task name: ",
+		    Pn, pid, tid);
+		safestrprt(tcmd, stderr, 1);
+		Exit(1);
+	    }
+	}
+#endif	/* defined(HASTASKS) */
+
 /*
  * Process the ID's current working directory info.
  */
-	(void) make_proc_path(idp, idpl, &path, &pathl, "cwd");
-	alloc_lfile(CWD, -1);
-	if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-	    if (!Fwarn) {
-		(void) memset((void *)&sb, 0, sizeof(sb));
-		lnk = ss = 0;
-		(void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
-		    strerror(errno));
-		nmabuf[sizeof(nmabuf) - 1] = '\0';
-		(void) add_nma(nmabuf, strlen(nmabuf));
-		pn = 1;
-	    } else
-		pn = 0;
-	} else {
-	    lnk = 1;
-	    ss = SB_ALL;
-	    if (HasNFS) {
-		if ((sv = statsafely(path, &sb)))
-		    sv = statEx(pbuf, &sb, &ss);
-	    } else
-		sv = stat(path, &sb);
-	    if (sv) {
-		ss = 0;
+	efs = 0;
+	if (!Ckscko) {
+	    (void) make_proc_path(idp, idpl, &path, &pathl, "cwd");
+	    alloc_lfile(CWD, -1);
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
 		if (!Fwarn) {
-		    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
-			strerror(errno));
-		    nmabuf[sizeof(nmabuf) - 1] = '\0';
-		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-	    }
-	    if (Scko) {
-		if ((ss & SB_MODE) && ((sb.st_mode & S_IFMT) == S_IFSOCK))
-		    pn = 1;
-		else
-		    pn = 0;
-	    } else
-		pn = 1;
-	}
-	if (pn) {
-	    (void) process_proc_node(lnk ? pbuf : path,
-				     &sb, ss,
-				     (struct stat *)NULL, 0);
-	    if (Lf->sf)
-		link_lfile();
-	}
-/*
- * Process the ID's root directory info.
- */
-	(void) make_proc_path(idp, idpl, &path, &pathl, "root");
-	alloc_lfile(RTD, -1);
-	if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-	    if (!Fwarn) {
-		(void) memset((void *)&sb, 0, sizeof(sb));
-		lnk = ss = 0;
-		(void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
-		    strerror(errno));
-		nmabuf[sizeof(nmabuf) - 1] = '\0';
-		(void) add_nma(nmabuf, strlen(nmabuf));
-		pn = 1;
-	    } else
-		pn = 0;
-	} else {
-	    lnk = 1;
-	    ss = SB_ALL;
-	    if (HasNFS) {
-		if ((sv = statsafely(path, &sb)))
-		    sv = statEx(pbuf, &sb, &ss);
-	    } else
-		sv = stat(path, &sb);
-	    if (sv) {
-		ss = 0;
-		if (!Fwarn) {
-		    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
-			strerror(errno));
-		    nmabuf[sizeof(nmabuf) - 1] = '\0';
-		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-	    }
-	    if (Scko) {
-		if ((ss & SB_MODE) && ((sb.st_mode & S_IFMT) == S_IFSOCK))
-		    pn = 1;
-		else
-		    pn = 0;
-	    } else
-		pn = 1;
-	}
-	if (pn) {
-	    (void) process_proc_node(lnk ? pbuf : path,
-				     &sb, ss,
-				     (struct stat *)NULL, 0);
-	    if (Lf->sf)
-		link_lfile();
-	}
-/*
- * Process the ID's execution info.
- */
-	(void) make_proc_path(idp, idpl, &path, &pathl, "exe");
-	alloc_lfile("txt", -1);
-	if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-	    (void) memset((void *)&sb, 0, sizeof(sb));
-	    lnk = ss = 0;
-	    if (!Fwarn) {
-		if ((errno != ENOENT) || uid) {
+		    zeromem((char *)&sb, sizeof(sb));
+		    lnk = ss = 0;
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
 			strerror(errno));
 		    nmabuf[sizeof(nmabuf) - 1] = '\0';
 		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-		pn = 1;
-	    } else
-		pn = 0;
-	 } else {
-	    lnk = 1;
-	    ss = SB_ALL;
-	    if (HasNFS) {
-		if ((sv = statsafely(path, &sb))) {
-		    sv = statEx(pbuf, &sb,  &ss);
-		    if (!sv && (ss & SB_DEV) && (ss & SB_INO))
-			txts = 1;
-		}
-	    } else
-		sv = stat(path, &sb);
-	    if (sv) {
-		ss = 0;
-		if (!Fwarn) {
-		    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
-			strerror(errno));
-		    nmabuf[sizeof(nmabuf) - 1] = '\0';
-		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-	    } else
-		txts = 1;
-	    if (Scko) {
-		if ((ss & SB_MODE) && ((sb.st_mode & S_IFMT) == S_IFSOCK))
 		    pn = 1;
-		else
+		} else
 		    pn = 0;
-	    } else
-		pn = 1;
-	}
-	if (pn) {
+	    } else {
+		lnk = pn = 1;
+		if (Efsysl && !isefsys(pbuf, "UNKNcwd", 1, NULL, &lfr)) {
+		    efs = 1;
+		    pn = 0;
+		} else {
+		    ss = SB_ALL;
+		    if (HasNFS) {
+			if ((sv = statsafely(path, &sb)))
+			sv = statEx(pbuf, &sb, &ss);
+		    } else
+			sv = stat(path, &sb);
+		    if (sv) {
+			ss = 0;
+			if (!Fwarn) {
+			    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
+				strerror(errno));
+			    nmabuf[sizeof(nmabuf) - 1] = '\0';
+			    (void) add_nma(nmabuf, strlen(nmabuf));
+			}
+		    }
+		}
+	    }
+	    if (pn) {
 		(void) process_proc_node(lnk ? pbuf : path,
-					 &sb, ss,
+					 path, &sb, ss,
 					 (struct stat *)NULL, 0);
 		if (Lf->sf)
 		    link_lfile();
+	    }
+	}
+/*
+ * Process the ID's root directory info.
+ */
+	lnk = ss = 0;
+	if (!Ckscko) {
+	    (void) make_proc_path(idp, idpl, &path, &pathl, "root");
+	    alloc_lfile(RTD, -1);
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
+		if (!Fwarn) {
+		    zeromem((char *)&sb, sizeof(sb));
+		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
+			strerror(errno));
+		    nmabuf[sizeof(nmabuf) - 1] = '\0';
+		    (void) add_nma(nmabuf, strlen(nmabuf));
+		    pn = 1;
+		} else
+		    pn = 0;
+	    } else {
+		lnk = pn = 1;
+		if (Efsysl && !isefsys(pbuf, "UNKNrtd", 1, NULL, NULL))
+		    pn = 0;
+		else {
+		    ss = SB_ALL;
+		    if (HasNFS) {
+			if ((sv = statsafely(path, &sb)))
+			    sv = statEx(pbuf, &sb, &ss);
+		    } else
+			sv = stat(path, &sb);
+		    if (sv) {
+			ss = 0;
+			if (!Fwarn) {
+			    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
+				strerror(errno));
+			    nmabuf[sizeof(nmabuf) - 1] = '\0';
+			    (void) add_nma(nmabuf, strlen(nmabuf));
+			}
+		    }
+		}
+	    }
+	    if (pn) {
+		(void) process_proc_node(lnk ? pbuf : path,
+					 path, &sb, ss,
+					 (struct stat *)NULL, 0);
+		if (Lf->sf)
+		    link_lfile();
+	    }
+	}
+/*
+ * Process the ID's execution info.
+ */
+	lnk = ss = txts = 0;
+	if (!Ckscko) {
+	    (void) make_proc_path(idp, idpl, &path, &pathl, "exe");
+	    alloc_lfile("txt", -1);
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
+		zeromem((void *)&sb, sizeof(sb));
+		if (!Fwarn) {
+		    if ((errno != ENOENT) || uid) {
+			(void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
+			    strerror(errno));
+			nmabuf[sizeof(nmabuf) - 1] = '\0';
+			(void) add_nma(nmabuf, strlen(nmabuf));
+		    }
+		    pn = 1;
+		} else
+		    pn = 0;
+	    } else {
+		lnk = pn = 1;
+		if (Efsysl && !isefsys(pbuf, "UNKNtxt", 1, NULL, NULL))
+		    pn = 0;
+		else {
+		    ss = SB_ALL;
+		    if (HasNFS) {
+			if ((sv = statsafely(path, &sb))) {
+			    sv = statEx(pbuf, &sb,  &ss);
+			    if (!sv && (ss & SB_DEV) && (ss & SB_INO))
+				txts = 1;
+			}
+		    } else
+			sv = stat(path, &sb);
+		    if (sv) {
+			ss = 0;
+			if (!Fwarn) {
+			    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
+				strerror(errno));
+			    nmabuf[sizeof(nmabuf) - 1] = '\0';
+			    (void) add_nma(nmabuf, strlen(nmabuf));
+			}
+		    } else
+			txts = 1;
+		}
+	    }
+	    if (pn) {
+		(void) process_proc_node(lnk ? pbuf : path,
+					 path, &sb, ss,
+					 (struct stat *)NULL, 0);
+		if (Lf->sf)
+		    link_lfile();
+	    }
 	}
 /*
  * Process the ID's memory map info.
  */
-	if (!Scko) {
+	if (!Ckscko) {
 	    (void) make_proc_path(idp, idpl, &path, &pathl, "maps");
 	    (void) process_proc_map(path, txts ? &sb : (struct stat *)NULL,
 				    txts ? ss : 0);
@@ -934,8 +1160,8 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
 		continue;
 	    (void) make_proc_path(dpath, i, &path, &pathl, fp->d_name);
 	    (void) alloc_lfile((char *)NULL, fd);
-	    if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-		(void) memset((void *)&sb, 0, sizeof(sb));
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), &rest) < 1) {
+		zeromem((char *)&sb, sizeof(sb));
 		lnk = ss = 0;
 		if (!Fwarn) {
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
@@ -947,69 +1173,104 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid)
 		    pn = 0;
 	    } else {
 		lnk = 1;
-		if (HasNFS) {
-		    if (lstatsafely(path, &lsb)) {
-			(void) statEx(pbuf, &lsb, &ls);
-		        enls = errno;
-		    } else {
-			enls = 0;
-			ls = SB_ALL;
-		    }
-		    if (statsafely(path, &sb)) {
-			(void) statEx(pbuf, &sb, &ss);
-			enss = errno;
-		    } else {
-			enss = 0;
-			ss = SB_ALL;
-		    }
+		if (Efsysl && !isefsys(pbuf, "UNKNfd", 1, NULL, &lfr)) {
+		    efs = 1;
+		    pn = 0;
 		} else {
-		    ls = lstat(path, &lsb) ? 0 : SB_ALL;
-		    enls = errno;
-		    ss = stat(path, &sb) ? 0 : SB_ALL;
-		    enss = errno;
-		}
-		if (!ls && !Fwarn) {
-		    (void) snpf(nmabuf, sizeof(nmabuf), "lstat: %s)",
-			strerror(enls));
-		    nmabuf[sizeof(nmabuf) - 1] = '\0';
-		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-		if (!ss && !Fwarn) {
-		    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
-			strerror(enss));
-		    nmabuf[sizeof(nmabuf) - 1] = '\0';
-		    (void) add_nma(nmabuf, strlen(nmabuf));
-		}
-		if (Scko) {
-		    if ((ss & SB_MODE) && ((sb.st_mode & S_IFMT) == S_IFSOCK))
+		    if (HasNFS) {
+			if (lstatsafely(path, &lsb)) {
+			    (void) statEx(pbuf, &lsb, &ls);
+			    enls = errno;
+			} else {
+			    enls = 0;
+			    ls = SB_ALL;
+			}
+			if (statsafely(path, &sb)) {
+			    (void) statEx(pbuf, &sb, &ss);
+			    enss = errno;
+			} else {
+			    enss = 0;
+			    ss = SB_ALL;
+			}
+		    } else {
+			ls = lstat(path, &lsb) ? 0 : SB_ALL;
+			enls = errno;
+			ss = stat(path, &sb) ? 0 : SB_ALL;
+			enss = errno;
+		    }
+		    if (!ls && !Fwarn) {
+			(void) snpf(nmabuf, sizeof(nmabuf), "lstat: %s)",
+			    strerror(enls));
+			nmabuf[sizeof(nmabuf) - 1] = '\0';
+			(void) add_nma(nmabuf, strlen(nmabuf));
+		    }
+		    if (!ss && !Fwarn) {
+			(void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
+			    strerror(enss));
+			nmabuf[sizeof(nmabuf) - 1] = '\0';
+			(void) add_nma(nmabuf, strlen(nmabuf));
+		    }
+		    if (Ckscko) {
+			if ((ss & SB_MODE)
+			&&  ((sb.st_mode & S_IFMT) == S_IFSOCK))
+			{
+			    pn = 1;
+			} else
+			    pn = 0;
+		    } else
 			pn = 1;
-		    else
-			pn = 0;
-		} else
-		    pn = 1;
+		}
 	    }
-	    if (pn) {
+	    if (pn || (efs && lfr && oty)) {
 		if (oty) {
 		    (void) make_proc_path(ipath, j, &pathi, &pathil,
 					  fp->d_name);
-		    if ((av = get_fdinfo(pathi, &fi)) & FDINFO_POS) {
-			ls |= SB_SIZE;
-			lsb.st_size = fi.pos;
+		    if ((av = get_fdinfo(pathi,FDINFO_ALL,&fi)) & FDINFO_POS) {
+			if (efs) {
+			    if (Foffset) {
+				lfr->off = (SZOFFTYPE)fi.pos;
+				lfr->off_def = 1;
+			    }
+			} else {
+			    ls |= SB_SIZE;
+			    lsb.st_size = fi.pos;
+			}
 		    } else
 			ls &= ~SB_SIZE;
 
 #if	!defined(HASNOFSFLAGS)
 		    if ((av & FDINFO_FLAGS) && (Fsv & FSV_FG)) {
-			Lf->ffg = (long)fi.flags;
-			Lf->fsv |= FSV_FG;
+			if (efs) {
+			    lfr->ffg = (long)fi.flags;
+			    lfr->fsv |= FSV_FG;
+			} else {
+			    Lf->ffg = (long)fi.flags;
+			    Lf->fsv |= FSV_FG;
+			}
 		     }
 # endif	/* !defined(HASNOFSFLAGS) */
 
 		}
-		process_proc_node(lnk ? pbuf : path,
-				  &sb, ss, &lsb, ls);
-		if (Lf->sf)
-		    link_lfile();
+		if (pn) {
+		    process_proc_node(lnk ? pbuf : path, path, &sb, ss, &lsb,
+				      ls);
+		    if ((Lf->ntype == N_ANON_INODE) && rest && *rest)
+			enter_nm(rest);
+
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+		    else if (Lf->rdev_def
+			 &&  is_pty_ptmx(Lf->rdev)
+			 &&  (av & FDINFO_TTY_INDEX)
+		    ) {
+			    enter_ptmxi(fi.tty_index);
+			    Lf->tty_index = fi.tty_index;
+			    Lf->sf |= SELPTYINFO;
+		    }
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
+		    if (Lf->sf)
+			link_lfile();
+		}
 	    }
 	}
 	(void) closedir(fdp);
@@ -1029,7 +1290,7 @@ process_proc_map(p, s, ss)
 {
 	char buf[MAXPATHLEN + 1], *ep, fmtbuf[32], **fp, nmabuf[MAXPATHLEN + 1];
 	dev_t dev;
-	int ds, en, i, mss, nf, sv;
+	int ds, efs, en, i, mss, nf, sv;
 	int eb = 6;
 	INODETYPE inode;
 	MALLOC_S len;
@@ -1042,6 +1303,7 @@ process_proc_map(p, s, ss)
 	    INODETYPE inode;
 	};
 	static struct saved_map *sm = (struct saved_map *)NULL;
+	efsys_list_t *rep;
 	static int sma = 0;
 	static char *vbuf = (char *)NULL;
 	static size_t vsz = (size_t)0;
@@ -1122,7 +1384,7 @@ process_proc_map(p, s, ss)
 		if (!sm) {
 		    (void) fprintf(stderr,
 			"%s: can't allocate %d bytes for saved maps, PID %d\n",
-			Pn, len, Lp->pid);
+			Pn, (int)len, Lp->pid);
 		    Exit(1);
 		}
 	    }
@@ -1130,30 +1392,38 @@ process_proc_map(p, s, ss)
 	    sm[ns++].inode = inode;
 	/*
 	 * Allocate space for the mapped file, then get stat(2) information
-	 * for it.
+	 * for it.  Skip the stat(2) operation if this is on an exempt file
+	 * system.
 	 */
 	    alloc_lfile("mem", -1);
-	    if (HasNFS) {
-		sv = statsafely(fp[6], &sb);
-	    } else
-		sv = stat(fp[6], &sb);
-	    if (sv) {
+	    if (Efsysl && !isefsys(fp[6], (char *)NULL, 0, &rep, NULL))
+		efs = sv = 1;
+	    else
+		efs = 0;
+	    if (!efs) {
+		if (HasNFS)
+		    sv = statsafely(fp[6], &sb);
+		else
+		    sv = stat(fp[6], &sb);
+	    }
+	    if (sv || efs) {
 		en = errno;
 	    /*
-	     * Applying stat(2) to the file failed, so manufacture a partial
+	     * Applying stat(2) to the file was not possible (file is on an
+	     * exempt file system) or stat(2) failed, so manufacture a partial
 	     * stat(2) reply from the process' maps file entry.
 	     *
-	     * If the file has been deleted, reset its type to "DEL"; otherwise
-	     * generate a stat() error name addition.
+	     * If the file has been deleted, reset its type to "DEL";
+	     * otherwise generate a stat() error name addition.
 	     */
-		(void) memset((void *)&sb, 0, sizeof(sb));
+		zeromem((char *)&sb, sizeof(sb));
 		sb.st_dev = dev;
 		sb.st_ino = (ino_t)inode;
 		sb.st_mode = S_IFREG;
 		mss = SB_DEV | SB_INO | SB_MODE;
 		if (ds)
 		    alloc_lfile("DEL", -1);
-		else {
+		else if (!efs && !Fwarn) {
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(stat: %s)",
 			strerror(en));
 		    nmabuf[sizeof(nmabuf) - 1] = '\0';
@@ -1195,7 +1465,7 @@ process_proc_map(p, s, ss)
 			(void) add_nma(nmabuf, strlen(nmabuf));
 		    }
 		}
-		(void) memset((void *)&sb, 0, sizeof(sb));
+		zeromem((char *)&sb, sizeof(sb));
 		sb.st_dev = dev;
 		sb.st_ino = (ino_t)inode;
 		sb.st_mode = S_IFREG;
@@ -1205,7 +1475,27 @@ process_proc_map(p, s, ss)
 	/*
 	 * Record the file's information.
 	 */
-	    process_proc_node(fp[6], &sb, mss, (struct stat *)NULL, 0);
+	    if (!efs)
+		process_proc_node(fp[6], fp[6], &sb, mss, (struct stat *)NULL,
+				  0);
+	    else {
+
+	    /*
+	     * If this file is on an exempt file system, complete the lfile
+	     * structure, but change its type and add the exemption note to
+	     * the NAME column.
+	     */
+		Lf->dev = sb.st_dev;
+		Lf->inode = (ino_t)sb.st_ino;
+		Lf->dev_def = Lf->inp_ty = 1;
+		(void) enter_nm(fp[6]);
+		(void) snpf(Lf->type, sizeof(Lf->type), "%s",
+			    (ds ? "UNKNdel" : "UNKNmem"));
+		(void) snpf(nmabuf, sizeof(nmabuf), "(%ce %s)",
+		    rep->rdlnk ? '+' : '-', rep->path);
+		nmabuf[sizeof(nmabuf) - 1] = '\0';
+		(void) add_nma(nmabuf, strlen(nmabuf));
+	    }
 	    if (Lf->sf)
 		link_lfile();
 	}
@@ -1216,14 +1506,14 @@ process_proc_map(p, s, ss)
 /*
  * read_id_stat() - read ID (PID or LWP ID) status
  *
- * return: -1 == ID is a zombie
+ * return: -1 == ID is unavailable
  *          0 == ID OK
- *          1 == ID unavailable
+ *          1 == ID is a zombie
+ *	    2 == ID is a thread
  */
 
 static int
-read_id_stat(ty, p, id, cmd, ppid, pgid)
-	int ty;				/* type: 0 == PID, 1 == LWP */
+read_id_stat(p, id, cmd, ppid, pgid)
 	char *p;			/* path to status file */
 	int id;				/* ID: PID or LWP */
 	char **cmd;			/* malloc'd command name */
@@ -1232,11 +1522,10 @@ read_id_stat(ty, p, id, cmd, ppid, pgid)
 					 * type */
 {
 	char buf[MAXPATHLEN], *cp, *cp1, **fp;
+	int ch, cx, es, nf, pc;
 	static char *cbf = (char *)NULL;
 	static MALLOC_S cbfa = 0;
 	FILE *fs;
-	MALLOC_S len;
-	int nf;
 	static char *vbuf = (char *)NULL;
 	static size_t vsz = (size_t)0;
 /*
@@ -1244,70 +1533,105 @@ read_id_stat(ty, p, id, cmd, ppid, pgid)
  * and read the file's first line.
  */
 	if (!(fs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
-	    return(1);
-	cp = fgets(buf, sizeof(buf), fs);
-	(void) fclose(fs);
-	if (!cp)
-	    return(1);
-/*
- * Separate the line into fields on white space separators.  Expect five fields
- * for a PID type and three for an LWP type.
- */
-	if ((nf = get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0))
-	<  (ty ? 5 : 3))
-	{
-	    return(1);
+	    return(-1);
+	if (!(cp = fgets(buf, sizeof(buf), fs))) {
+
+read_id_stat_exit:
+
+	    (void) fclose(fs);
+	    return(-1);
 	}
 /*
- * Convert the first field to an integer; its conversion must match the
- * ID argument.
+ * Skip to the first field, and make sure it is a matching ID.
  */
-	if (!fp[0] || (atoi(fp[0]) != id))
-	    return(1);
-/*
- * Get the command name from the second field.  Strip a starting '(' and
- * an ending ')'.  Allocate space to hold the result and return the space
- * pointer.
- */
-	if (!(cp = fp[1]))
-	    return(1);
-	if (cp && *cp == '(')
+	cp1 = cp;
+	while (*cp && (*cp != ' ') && (*cp != '\t'))
 	    cp++;
-	if ((cp1 = strrchr(cp, ')')))
-	    *cp1 = '\0';
-	if ((len = strlen(cp) + 1) > cbfa) {
-	     cbfa = len;
-	     if (cbf)
-		cbf = (char *)realloc((MALLOC_P *)cbf, cbfa);
-	     else
-		cbf = (char *)malloc(cbfa);
-	     if (!cbf) {
-		(void) fprintf(stderr,
-		    "%s: can't allocate %d bytes for command \"%s\"\n",
-		    Pn, cbfa, cp);
-		Exit(1);
-	     }
+	if (*cp)
+	    *cp = '\0';
+	if (atoi(cp1) != id)
+	    goto read_id_stat_exit;
+/*
+ * The second field should contain the command, enclosed in parentheses.
+ * If it also has embedded '\n' characters, replace them with '?' characters,
+ * accumulating command characters until a closing parentheses appears.
+ *
+ */
+	for (++cp; *cp && (*cp == ' '); cp++)
+		;
+	if (!cp || (*cp != '('))
+	    goto read_id_stat_exit;
+	cp++;
+	pc = 1;			/* start the parenthesis balance count at 1 */
+/*
+ * Enter the command characters safely.  Supply them from the initial read
+ * of the stat file line, a '\n' if the initial read didn't yield a ')'
+ * command closure, or by reading the rest of the command a character at
+ * a time from the stat file.  Count embedded '(' characters and balance
+ * them with embedded ')' characters.  The opening '(' starts the balance
+ * count at one.
+ */
+	for (cx = es = 0;;) {
+	    if (!es)
+		ch = *cp++;
+	    else {
+		if ((ch = fgetc(fs)) == EOF)
+		    goto read_id_stat_exit;
+	    }
+	    if (ch == '(')		/* a '(' advances the balance count */
+		pc++;
+	    if (ch == ')') {
+	    
+	    /*
+	     * Balance parentheses when a closure is encountered.  When
+	     * they are balanced, this is the end of the command.
+	     */
+		pc--;
+		if (!pc)
+		    break;
+	    }
+	    if ((cx + 2) > cbfa)
+		cbfa = alloc_cbf((cx + 2), &cbf, cbfa);
+	    cbf[cx] = ch;
+	    cx++;
+	    cbf[cx] = '\0';
+	    if (!es && !*cp)
+		es = 1;		/* Switch to fgetc() when a '\0' appears. */
 	}
-	(void) snpf(cbf, len, "%s", cp);
 	*cmd = cbf;
 /*
- * If the type is PID, convert and return parent process (fourth field)
- * and process group (fifth field) IDs.
+ * Read the remainder of the stat line if it was necessary to read command
+ * characters individually from the stat file.
+ *
+ * Separate the reminder into fields.
  */
-	if (!ty) {
-	    if (fp[3] && *fp[3])
-		*ppid = atoi(fp[3]);
-	    else
-		return(1);
-	    if (fp[4] && *fp[4])
-		*pgid = atoi(fp[4]);
-	    else
-		return(1);
-	}
+	if (es)
+	    cp = fgets(buf, sizeof(buf), fs);
+	(void) fclose(fs);
+	if (!cp || !*cp)
+	    return(-1);
+	if ((nf = get_fields(cp, (char *)NULL, &fp, (int *)NULL, 0)) < 3)
+	    return(-1);
+/*
+ * Convert and return parent process (fourth field) and process group (fifth
+ * field) IDs.
+ */
+	if (fp[1] && *fp[1])
+	    *ppid = atoi(fp[1]);
+	else
+	    return(-1);
+	if (fp[2] && *fp[2])
+	    *pgid = atoi(fp[2]);
+	else
+	    return(-1);
 /*
  * Check the state in the third field.  If it is 'Z', return that indication.
  */
-	return((fp[2] && !strcmp(fp[2], "Z")) ? -1 : 0);
+	if (fp[0] && !strcmp(fp[0], "Z"))
+	    return(1);
+	else if (fp[0] && !strcmp(fp[0], "T"))
+	    return(2);
+	return(0);
 }
 
 
@@ -1357,7 +1681,7 @@ statEx(p, s, ss)
 	}
 	(void) strcpy(cb, p);
 /*
- * Trim trailing leaves from the end of the path one at a time and do s safe
+ * Trim trailing leaves from the end of the path one at a time and do a safe
  * stat() on each trimmed result.  Stop when a safe stat() succeeds or doesn't
  * fail because of EACCES or EPERM.
  */
@@ -1376,7 +1700,7 @@ statEx(p, s, ss)
  * If a stat() on a trimmed result succeeded, form partial results containing
  * only the device and raw device numbers.
  */
-	memset((void *)s, 0, sizeof(struct stat));
+	zeromem((char *)s, sizeof(struct stat));
 	if (st) {
 	    errno = 0;
 	    s->st_dev = sb.st_dev;
